#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include "cJSON.h"
#include "semaphore.h"

#include "log_util.h"
#include "iot_export.h"
#include "iot_export_linkkit.h"

#define USER_EXAMPLE_YIELD_TIMEOUT_MS (200)

typedef struct {
    int master_devid;
    int cloud_connected;
    int master_initialized;
} user_example_ctx_t;

static user_example_ctx_t g_user_example_ctx;

static user_example_ctx_t *user_example_get_ctx(void) {
    return &g_user_example_ctx;
}

static int user_disconnected_event_handler(void) {
    user_example_ctx_t *user_example_ctx = user_example_get_ctx();

    LOG_INFO("Cloud Disconnected.\n");

    user_example_ctx->cloud_connected = 0;

    return 0;
}


static int user_connected_event_handler(void) {
    user_example_ctx_t *user_example_ctx = user_example_get_ctx();

    LOG_INFO("Cloud Connected.\n");
    user_example_ctx->cloud_connected = 1;
    return 0;
}

#ifdef __APPLE__
static sem_t *g_sem_linkkit_loop;
#else
static sem_t g_sem_linkkit_loop;
#endif
static int g_flag_is_running = 1;

void *linkkit_loop(void *arg) {
#ifdef __APPLE__
    sem_post(g_sem_linkkit_loop);
#else
    sem_post(&g_sem_linkkit_loop);
#endif

    while (1 == g_flag_is_running) {
        IOT_Linkkit_Yield(USER_EXAMPLE_YIELD_TIMEOUT_MS);
    }
    LOG_INFO("mqtt client exit.\n");

#ifdef __APPLE__
    sem_post(g_sem_linkkit_loop);
#else
    sem_post(&g_sem_linkkit_loop);
#endif
    return NULL;
}

int linkface_service_adapter_request_event_handler(const int devid, const char *service_id, const int service_id_len,
                                                   const char *service_name, const int service_name_len,
                                                   const char *request,
                                                   const int request_len, char **response, int *response_len) {
    int ret = 0;
    LOG_INFO(
            "Service Request Received, devid: %d, service_id: %.*s, service_name: %.*s, request len: %d, request: %.*s \n",
            devid, service_id_len, service_id, service_name_len, service_name, request_len, request_len, request);

    //如果ret的返回值小于0，表明sdk不支持此种服务调用，可能是厂商自定义的物模型，需要厂商自行处理。
    if (ret < 0) {
        //todo.
        LOG_INFO("linkface return code: %d\n", ret);
    }

    return ret;
}

int linkkit_init_adapter(const char *product_key, const char *device_name, const char *device_secret) {
    LOG_INFO("linkkit adapter init...\n");
    int res = 0;
    iotx_linkkit_dev_meta_info_t master_meta_info;
    user_example_ctx_t *user_example_ctx = user_example_get_ctx();

    if ((NULL == product_key) || (NULL == device_name) || (NULL == device_secret)) {
        LOG_INFO("init adapter input param was Null.\n");
        return -1;
    }

    memset(user_example_ctx, 0, sizeof(user_example_ctx_t));

    IOT_SetLogLevel(IOT_LOG_ERROR);

    /* Register Callback */
    IOT_RegisterCallback(ITE_CONNECT_SUCC, user_connected_event_handler);
    IOT_RegisterCallback(ITE_DISCONNECTED, user_disconnected_event_handler);
    IOT_RegisterCallback(ITE_SERVICE_REQUST, linkface_service_adapter_request_event_handler);

    memset(&master_meta_info, 0, sizeof(iotx_linkkit_dev_meta_info_t));
    strcpy(master_meta_info.product_key, product_key);
    strcpy(master_meta_info.device_name, device_name);
    strcpy(master_meta_info.device_secret, device_secret);

    // 选择站点
    int domain_type = IOTX_CLOUD_REGION_SHANGHAI;
    IOT_Ioctl(IOTX_IOCTL_SET_DOMAIN, (void *) &domain_type);

    // 选择登录方式
    int dynamic_register = 0;
    IOT_Ioctl(IOTX_IOCTL_SET_DYNAMIC_REGISTER, (void *) &dynamic_register);

    /* Choose Whether You Need Post Property/Event Reply */
    int post_event_reply = 1;
    IOT_Ioctl(IOTX_IOCTL_RECV_EVENT_REPLY, (void *) &post_event_reply);

    /* Create Master Device Resources */
    user_example_ctx->master_devid = IOT_Linkkit_Open(IOTX_LINKKIT_DEV_TYPE_MASTER, &master_meta_info);
    if (user_example_ctx->master_devid < 0) {
        LOG_INFO("IOT_Linkkit_Open Failed\n");
        return -1;
    }

    /* Start Connect Aliyun Server */
    res = IOT_Linkkit_Connect(user_example_ctx->master_devid);
    if (res < 0) {
        IOT_Linkkit_Close(user_example_ctx->master_devid);
        LOG_INFO("IOT_Linkkit_Connect Failed\n");
        return -1;
    }

    pthread_t thread_id = 0;

#ifdef __APPLE__
    g_sem_linkkit_loop = sem_open("sem_linkkit_loop", O_CREAT);
#else
    sem_init(&g_sem_linkkit_loop, 0, 0);
#endif
    g_flag_is_running = 1;
    pthread_create(&thread_id, NULL, linkkit_loop, NULL);
    pthread_detach(thread_id);
    if (thread_id > 0) {
        //必须等到linkkit_loop线程起来之后才返回。
#ifdef __APPLE__
        sem_wait(g_sem_linkkit_loop);
#else
        sem_wait(&g_sem_linkkit_loop);
#endif
        return 1;
    } else {

#ifdef __APPLE__
        sem_close(g_sem_linkkit_loop);
#else
        sem_destroy(&g_sem_linkkit_loop);
#endif
        IOT_Linkkit_Close(user_example_ctx->master_devid);
        return -1;
    }
}

int linkkit_destroy_adapter() {
    user_example_ctx_t *user_example_ctx = user_example_get_ctx();
    g_flag_is_running = 0;
    sleep(2);
    IOT_Linkkit_Close(user_example_ctx->master_devid);
    LOG_INFO("linkkit adapter destroy done.\n");
    return 0;
}

